#!/usr/bin/env python
"""
A Python implementation of rsync

This is a demonstration implementation of the rsync algorithm in Python. It is 
not fast and is not optimised. The primary aim is to provide a simple example
implementation of the algorithm for reference, so code clarity is more important
than performance. Ideas have been liberaly taken from libhsync, xdelta and 
rsync.

     $Id: pysync.py,v 0.1 2000/11/10 11:01:08 abo Exp $
Author	: Donovan Baarda <abo@minkirri.apana.org.au>
License	: GPL
Download: ftp://minkirri.apana.org.au/pub/python

Requires: sys, zlib, types, md5sum
    adler32 (included md5sum-alike adler32 wrapper)

Usage:
    # to create a signature of a file
    oldfile=open("someoldfilename",'rb')
    sig=calcsig(oldfile)
    oldfile.close()
    
    # to create a delta to a file using a signature
    newfile=open("somenewfilename",'rb')
    delta=calcdelta(newfile,sig)
    newfile.close()
    
    # to apply a delta to an old file to create a new file
    oldfile=open("someoldfilename",'rb')
    newfile=open("somenewfilename",'wb')
    applydelta(oldfile,delta,newfile)
    oldfile.close()
    newfile.close()

Where:
    oldfile     - the source file
    newfile     - the target file
    sigfile     - the signature file
    diffile     - the delta file


thoughts on using zlib to compress deltas;

1) compress the whole instruction stream
2) compress the inserts only using Z_SYNC_FLUSH to delimit and put
   inserts into the instruction stream.
3) compress everything using Z_SYNC_FLUSH to delimit boundaries, inserting
   only output for inserts into the instruction stream (rsync?)
4) compress the insert stream without Z_SYNC_FLUSH and put offset/lengths in
   instruction stream, sending compressed inserts seperately (xdelta?)

it depends on how zlib performs with heaps of Z_SYNC_FLUSH's. If it hurts
performance badly, then 4 is best. Otherwise, it would pay to see if zlib
improves compression with inserted context data not included in the output
stream.

My tests on zlib suggest that syncs do hurt a little, but dispite that
including context by compressing _all_ the data, not just the deltas, gives 
the best compression. Unfortunately this has extra load on applying patches 
because it requires all data to be compressed to supply the compression 
stream for the missing context info for decompression.

thoughts on instruction stream;

use fixed length and put only offsets into instruction stream for matches,
put inserts directly into the instruction stream.

use source/offset/length in the instruction stream, and make the inserts a
seperate source (xdelta).

by putting offset/length in the instruction stream rather than just block id's
the instruction stream becomes more generic... anything that can generate 
offset/lengths can generate patches... possibly more optimal ones than rsync
(ie, xdelta's larges possible match type tricks). 

Including a source along with offset/length means multiple sources can be used
for a single patch (like xdelta), though this can be fudged by appending sources
into one long stream.

a delta is implemented as a list containing a sequence of (context) compressed
insert strings and (offset,length) match pairs.

A signature is a (length, blocksize, sigtable) tuple, where length and blocksize
are integers. The sigtable is implemented as a adler32 keyed dictionary of
md5sum keyed dictionaries containing offsets.
ie sigtable[adler32][md5sum]=offset

Note rsync uses more compact md4sums, but python doesn't have a built in md4sum
wrapper, so I use md5sum.

TODO:
I'm still not happy with how clean this is... it could be clearer. The general
API is just slapped together with little thought about how it could be done
neatly. For clarity I'd probably prefer if it operated at the lowest level on
raw data than file objects... it needs more thought.

"""

from zlib import *
from types import TupleType
import md5,adler32

# the default block size used throughout. This is carefuly chosen to try and
# avoid the zlib decompressor sync bug which strikes at about 16K
BLOCK_SIZE=8192

def calcsig(oldfile,blocksize=BLOCK_SIZE):
    offset=0
    sigtable={}
    data=oldfile.read(blocksize)
    while data:
        sum=md5.new(data).digest()
        sig=adler32.new(data).digest()
        try:
            sigtable[sig][sum]=offset
        except KeyError:
            sigtable[sig]={}
            sigtable[sig][sum]=offset
        offset=offset+len(data)
        data=oldfile.read(blocksize)
    return (offset,blocksize,sigtable)

def sigstats((length,blocksize,sigtable)):
    keys=len(sigtable.keys())
    blks=(length+blocksize-1)/blocksize
    print "signature stats"
    print "length             :",length,blocksize,blks
    print "adler32 keys       :",keys
    print "adler32 collisions :",blks - keys


def calcdelta(newfile,(length,blocksize,sigtable)):
    delta=[]
    miss_insert=''
    match_offset=match_length=-1
    cmp=compressobj(9)
    data=newfile.read(blocksize)
    sig=adler32.new(data)
    while data:
        try:
            # get the matching offset, if it exists, otherwise raise KeyError
            sigmatch=sigtable[sig.digest()]
            sum=md5.new(data).digest()
            offset=sigmatch[sum]
            # if there was missed data before this match, append it to delta first
            if miss_insert:
                delta.append(cmp.compress(miss_insert)+cmp.flush(Z_SYNC_FLUSH))
                miss_insert=''
            # if the last block was matched, extend it
            if match_offset>=0:
                match_length=match_length+len(data)
            # else record the match offset and length
            else:
                match_offset,match_length=offset,len(data)
            # compress the matched block for context info
            cmp.compress(data)
            # get next block of data and calculate the adler32 sig
            data=newfile.read(blocksize)
            sig=adler32.new(data)
        except KeyError:
            # if there was a match before this miss, append it to delta first
            if match_offset>=0:
                delta.append((match_offset,match_length))
                cmp.flush(Z_SYNC_FLUSH)
                match_offset=-1
            # get the next input byte
            xn=newfile.read(1)
            # if not eof, append to miss_insert and rotate the adler32 sig
            if xn:
                x1,data=data[0],data[1:]+xn
                miss_insert=miss_insert+x1
                sig.rotate(x1,xn)
            # else the last block of data is an insert
            else:
                miss_insert,data=miss_insert+data,''
                delta.append(cmp.compress(miss_insert)+cmp.flush())
    # if there is a remaining match, append it to delta
    if match_offset>=0:
        delta.append((match_offset,match_length))
        cmp.flush()
    return delta    
              
def deltastats(delta):
    matches=inserts=match_length=insert_length=0
    for i in delta:
        if type(i)==TupleType:
            matches=matches+1
            match_length=match_length+i[1]
        else:
            inserts=inserts+1
            insert_length=insert_length + len(i)
    print "delta stats"
    print "segments:",len(delta)
    print "matches :",matches,match_length
    print "inserts :",inserts,insert_length

              
def applydelta(oldfile,delta,newfile):
    cmp=compressobj(9)
    out=decompressobj()
    for i in delta:
        if type(i)==TupleType:
            match_offset,match_length=i
            oldfile.seek(match_offset)
            # read the matched data in chunks
            while match_length:
                data=oldfile.read(min(match_length,BLOCK_SIZE))
                newfile.write(data)
                # compress the context info and feed it to the decompressor
                out.decompress(cmp.compress(data))
                match_length=match_length-len(data)
            # decompress a sync-flush after the match
            out.decompress(cmp.flush(Z_SYNC_FLUSH))
        else:
            miss_insert=i
            # decompress the insert data in chunks
            while miss_insert:
                data=out.decompress(miss_insert[:BLOCK_SIZE])
                newfile.write(data)
                # re-compress the insert data for context info
                cmp.compress(data)
                miss_insert=miss_insert[BLOCK_SIZE:]
            # sync_flush the compressor after the insert
            cmp.flush(Z_SYNC_FLUSH)


if __name__ == "__main__":
    import cPickle
    from sys import argv
    if len(argv)==4 and argv[1]=="sig":
        oldfile=open(argv[2],'rb')
        sigfile=open(argv[3],'wb')
        sig=calcsig(oldfile,1024)
        cPickle.dump(sig,sigfile,1)
        oldfile.close()
        sigfile.close()
        sigstats(sig)
    elif len(argv)==5 and argv[1]=="delta":
        newfile=open(argv[2],'rb')
        sigfile=open(argv[3],'rb')
        diffile=open(argv[4],'wb')
        sig=cPickle.load(sigfile)
        delta=calcdelta(newfile,sig)
        cPickle.dump(delta,diffile,1)
        newfile.close()
        sigfile.close()
        diffile.close()
        deltastats(delta)
    elif len(argv)==5 and argv[1]=="apply":
        oldfile=open(argv[2],'rb')
        diffile=open(argv[3],'rb')
        newfile=open(argv[4],'wb')
        delta=cPickle.load(diffile)
        applydelta(oldfile,delta,newfile)
        oldfile.close()
        diffile.close()
        newfile.close()
    else:
        print """
Usage:
    pysync sig <oldfile> <sigfile>
    	... generates signature file <sigfile> from <oldfile>
    
    pysync delta <newfile> <sigfile> <diffile>
        ... generates delta file <diffile> for <newfile> from <sigfile>
        
    pysync apply <oldfile> <diffile> <newfile>
        ... applies delta file <diffile> to <oldfile> to generate <newfile>
"""    
